from __future__ import (
    annotations,
)

import asyncio
import time
from asyncio import (
    gather,
    iscoroutinefunction,
)
from collections.abc import (
    Awaitable,
    Callable,
    Iterable,
)
from functools import (
    wraps,
)
from inspect import (
    isawaitable,
)
from typing import (
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from ...exceptions import (
    NotSatisfiedCheckerException,
)
from ...requests import (
    Request,
)

Checker = Callable[[Request], Union[Optional[bool], Awaitable[Optional[bool]]]]


@runtime_checkable
class CheckerWrapper(Protocol):
    """Checker Wrapper class."""

    meta: CheckerMeta
    __call__: Checker


class CheckerMeta:
    """Checker Meta class."""

    func: Checker
    max_attempts: int
    delay: float

    def __init__(self, func: Checker, max_attempts: int, delay: float):
        self.func = func
        self.max_attempts = max_attempts
        self.delay = delay

    @staticmethod
    async def run_async(metas: set[CheckerMeta], *args, **kwargs) -> None:
        """Run a set of checkers asynchronously.

        :param metas: The set of checkers.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        metas = list(metas)
        futures = [meta.async_wrapper(*args, **kwargs) for meta in metas]

        for satisfied, meta in zip(await gather(*futures), metas):
            if not satisfied:
                raise NotSatisfiedCheckerException(f"{meta.wrapper!r} is not satisfied.")

    @staticmethod
    def run_sync(metas: set[CheckerMeta], *args, **kwargs) -> bool:
        """Run a set of checkers synchronously.

        :param metas: The set of checkers.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        for meta in metas:
            satisfied = meta.sync_wrapper(*args, **kwargs)
            if not satisfied:
                raise NotSatisfiedCheckerException(f"{meta.wrapper!r} is not satisfied.")
        return True

    @property
    def wrapper(self) -> CheckerWrapper:
        """Get the ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """
        if iscoroutinefunction(self.func):
            return self.async_wrapper
        else:
            return self.sync_wrapper

    @cached_property
    def async_wrapper(self) -> CheckerWrapper:
        """Get the async ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """

        @wraps(self.func)
        async def _wrapper(*args, **kwargs) -> bool:
            r = 0
            while r < self.max_attempts:
                satisfied = self.func(*args, **kwargs)
                if isawaitable(satisfied):
                    satisfied = await satisfied
                if satisfied:
                    return True

                await asyncio.sleep(self.delay)
                r += 1

            return False

        _wrapper.meta = self
        return _wrapper

    @cached_property
    def sync_wrapper(self) -> CheckerWrapper:
        """Get the sync ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """
        if iscoroutinefunction(self.func):
            raise ValueError(f"{self.func!r} cannot be awaitable.")

        @wraps(self.func)
        def _wrapper(*args, **kwargs) -> bool:
            r = 0
            while r < self.max_attempts:
                satisfied = self.func(*args, **kwargs)
                if satisfied:
                    return True

                time.sleep(self.delay)
                r += 1

            return False

        _wrapper.meta = self
        return _wrapper

    def __repr__(self):
        args = ", ".join(map(repr, self))
        return f"{type(self).__name__}({args})"

    def __eq__(self, other: CheckerMeta) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        yield from (
            self.func,
            self.max_attempts,
            self.delay,
        )



from __future__ import (
    annotations,
)

from minos.common import (
    MinosConfig,
)

from .kafka import (
    KafkaBrokerPublisher,
)
from .queued import (
    InMemoryBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueue,
    QueuedBrokerPublisher,
)


class PostgreSqlQueuedKafkaBrokerPublisher(QueuedBrokerPublisher):
    """PostgreSql Queued Kafka Broker Publisher class."""

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> PostgreSqlQueuedKafkaBrokerPublisher:
        impl = KafkaBrokerPublisher.from_config(config, **kwargs)
        queue = PostgreSqlBrokerPublisherQueue.from_config(config, **kwargs)
        return cls(impl, queue, **kwargs)


class InMemoryQueuedKafkaBrokerPublisher(QueuedBrokerPublisher):
    """In Memory Queued Kafka Broker Publisher class."""

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> InMemoryQueuedKafkaBrokerPublisher:
        impl = KafkaBrokerPublisher.from_config(config, **kwargs)
        queue = InMemoryBrokerPublisherQueue.from_config(config, **kwargs)
        return cls(impl, queue, **kwargs)



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    TimeoutError,
    wait_for,
)
from contextlib import (
    suppress,
)

from aiokafka import (
    AIOKafkaProducer,
)
from cached_property import (
    cached_property,
)

from minos.common import (
    MinosConfig,
)

from ..messages import (
    BrokerMessage,
)
from .abc import (
    BrokerPublisher,
)

logger = logging.getLogger(__name__)


class KafkaBrokerPublisher(BrokerPublisher):
    """Kafka Broker Publisher class."""

    def __init__(self, *args, broker_host: str, broker_port: int, **kwargs):
        super().__init__(*args, **kwargs)
        self.broker_host = broker_host
        self.broker_port = broker_port

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> KafkaBrokerPublisher:
        kwargs["broker_host"] = config.broker.host
        kwargs["broker_port"] = config.broker.port
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        await self.client.start()

    async def _destroy(self) -> None:
        with suppress(TimeoutError):
            await wait_for(self.client.stop(), 0.5)
        await super()._destroy()

    async def _send(self, message: BrokerMessage) -> None:
        await self.client.send_and_wait(message.topic, message.avro_bytes)

    @cached_property
    def client(self) -> AIOKafkaProducer:
        """Get the client instance.

        :return: An ``AIOKafkaProducer`` instance.
        """
        return AIOKafkaProducer(bootstrap_servers=f"{self.broker_host}:{self.broker_port}")



from .abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)
from .memory import (
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
)
from .pg import (
    PostgreSqlBrokerSubscriberQueue,
    PostgreSqlBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueueQueryFactory,
)



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
)
from collections.abc import (
    Iterable,
)
from typing import (
    TypeVar,
)

from .....utils import (
    Builder,
)
from ....collections import (
    BrokerQueue,
)

logger = logging.getLogger(__name__)


class BrokerSubscriberQueue(BrokerQueue, ABC):
    """Broker Subscriber Queue class."""

    def __init__(self, topics: Iterable[str], **kwargs):
        super().__init__(**kwargs)
        topics = set(topics)
        if not len(topics):
            raise ValueError("The topics set must not be empty.")
        self._topics = topics

    @property
    def topics(self) -> set[str]:
        """Topics getter.

        :return: A list of string values.
        """
        return self._topics


class BrokerSubscriberQueueBuilder(Builder[BrokerSubscriberQueue], ABC):
    """Broker Subscriber Queue Builder class."""

    def with_topics(self: B, topics: Iterable[str]) -> B:
        """Set topics.

        :param topics: The topics to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["topics"] = set(topics)
        return self


B = TypeVar("B", bound=BrokerSubscriberQueueBuilder)



from __future__ import (
    annotations,
)

import logging
from typing import (
    Optional,
)

from ....collections import (
    PostgreSqlBrokerQueue,
    PostgreSqlBrokerQueueQueryFactory,
)
from .abc import (
    BrokerPublisherQueue,
)

logger = logging.getLogger(__name__)


class PostgreSqlBrokerPublisherQueue(PostgreSqlBrokerQueue, BrokerPublisherQueue):
    """PostgreSql Broker Publisher Queue class."""

    def __init__(self, *args, query_factory: Optional[PostgreSqlBrokerQueueQueryFactory] = None, **kwargs):
        if query_factory is None:
            query_factory = PostgreSqlBrokerPublisherQueueQueryFactory()
        super().__init__(*args, query_factory=query_factory, **kwargs)


class PostgreSqlBrokerPublisherQueueQueryFactory(PostgreSqlBrokerQueueQueryFactory):
    """PostgreSql Broker Publisher Queue Query Factory class."""

    def build_table_name(self) -> str:
        """Get the table name.

        :return: A ``str`` value.
        """
        return "broker_publisher_queue"



from __future__ import (
    annotations,
)

import logging
import warnings
from enum import (
    Enum,
)
from functools import (
    total_ordering,
)
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
    uuid4,
)

from minos.common import (
    DeclarativeModel,
)

from .abc import (
    BrokerMessage,
)

logger = logging.getLogger(__name__)


@total_ordering
class BrokerMessageV1(BrokerMessage, DeclarativeModel):
    """Broker Message V1 class."""

    topic: str
    identifier: UUID
    reply_topic: Optional[str]
    strategy: BrokerMessageV1Strategy  # FIXME: Remove this attribute!

    payload: BrokerMessageV1Payload

    def __init__(
        self,
        topic: str,
        payload: BrokerMessageV1Payload,
        *,
        identifier: Optional[UUID] = None,
        strategy: Optional[BrokerMessageV1Strategy] = None,
        **kwargs,
    ):
        if identifier is None:
            identifier = uuid4()
        if strategy is None:
            strategy = BrokerMessageV1Strategy.UNICAST

        super().__init__(topic=topic, identifier=identifier, strategy=strategy, payload=payload, **kwargs)

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def version(cls) -> int:
        """Get the version of the message.

        :return: A strictly positive ``int`` value.
        """
        return 1

    @property
    def topic(self) -> str:
        """Get the topic of the message.

        :return: A ``str`` value.
        """
        return self.fields["topic"].value

    @property
    def identifier(self) -> UUID:
        """Get the identifier of the message.

        :return: An ``UUID`` instance.
        """
        return self.fields["identifier"].value

    @property
    def reply_topic(self) -> Optional[str]:
        """Get the reply topic of the message if there is someone.

        :return: A ``str`` value or ``None``.
        """
        return self.fields["reply_topic"].value

    def set_reply_topic(self, value: Optional[str]) -> None:
        """Set the message's reply topic.

        :param value: A ``str`` value or ``None``.
        :return: This method does not return anything.
        """
        self.fields["reply_topic"].value = value

    @property
    def content(self) -> Any:
        """Get the content of the message.

        :return: Any value.
        """
        return self.payload.content

    @property
    def ok(self) -> bool:
        """Check if the message is okay or not.

        :return: ``True`` if the message is okay or ``False`` otherwise.
        """
        return self.payload.ok

    @property
    def status(self) -> int:
        """Get the payload status.

        :return: A ``BrokerMessageV1Status`` instance.
        """
        return self.payload.status

    @property
    def headers(self) -> dict[str, str]:
        """Get the payload headers.

        :return: A `dict` with `str` keys and `str` values.
        """
        return self.payload.headers

    @property
    def data(self) -> Any:
        """Get the payload content.

        :return: Any value.
        """
        warnings.warn("The `BrokerMessage.data` attribute has been deprecated", DeprecationWarning)
        return self.payload.content

    def __lt__(self, other: Any) -> bool:
        # noinspection PyBroadException
        return isinstance(other, type(self)) and self.payload < other.payload


@total_ordering
class BrokerMessageV1Payload(DeclarativeModel):
    """Broker Message Payload class."""

    content: Any
    status: BrokerMessageV1Status
    headers: dict[str, str]

    def __init__(
        self, content: Any = None, headers: Optional[dict[str, str]] = None, status: Optional[int] = None, **kwargs
    ):
        if headers is None:
            headers = dict()
        if status is None:
            status = BrokerMessageV1Status.SUCCESS
        super().__init__(content=content, status=status, headers=headers, **kwargs)

    @property
    def ok(self) -> bool:
        """Check if the message is okay or not.

        :return: ``True`` if the message is okay or ``False`` otherwise.
        """
        return self.status == BrokerMessageV1Status.SUCCESS

    @property
    def data(self) -> Any:
        """Get the content.

        :return: Any value.
        """
        warnings.warn("The `BrokerMessage.data` attribute has been deprecated", DeprecationWarning)
        return self.content

    def __lt__(self, other: Any) -> bool:
        # noinspection PyBroadException
        try:
            return isinstance(other, type(self)) and self.content < other.content
        except Exception:
            return False


class BrokerMessageV1Status(int, Enum):
    """Broker Message Status v1 class."""

    SUCCESS = 200
    ERROR = 400
    SYSTEM_ERROR = 500
    UNKNOWN = 600

    @classmethod
    def _missing_(cls, value: Any) -> BrokerMessageV1Status:
        logger.warning(f"The given status value is unknown: {value}")
        return cls.UNKNOWN


class BrokerMessageV1Strategy(str, Enum):
    """Broker Message Strategy class"""

    UNICAST = "unicast"
    MULTICAST = "multicast"



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    AsyncIterator,
    Iterable,
)
from typing import (
    Optional,
)

from minos.common import (
    MinosSetup,
)

from ...utils import (
    Builder,
)
from ..messages import (
    BrokerMessage,
)

logger = logging.getLogger(__name__)


class BrokerSubscriber(ABC, MinosSetup):
    """Broker Subscriber class."""

    def __init__(self, topics: Iterable[str], **kwargs):
        super().__init__(**kwargs)
        self._topics = set(topics)

    @property
    def topics(self) -> set[str]:
        """Topics getter.

        :return: A list of string values.
        """
        return self._topics

    def __aiter__(self) -> AsyncIterator[BrokerMessage]:
        return self

    async def __anext__(self) -> BrokerMessage:
        if self.already_destroyed:
            raise StopAsyncIteration
        return await self.receive()

    async def receive(self) -> BrokerMessage:
        """Receive a new message.

        :return: A ``BrokerMessage`` instance.
        """
        message = await self._receive()
        logger.info(f"Receiving {message!r} message...")
        return message

    @abstractmethod
    async def _receive(self) -> BrokerMessage:
        raise NotImplementedError


class BrokerSubscriberBuilder(Builder[BrokerSubscriber], ABC):
    """Broker Subscriber Builder class."""

    def with_group_id(self, group_id: Optional[str]):
        """Set group_id.

        :param group_id: The group_id to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["group_id"] = group_id
        return self

    def with_remove_topics_on_destroy(self, remove_topics_on_destroy: bool):
        """Set remove_topics_on_destroy.

        :param remove_topics_on_destroy: The remove_topics_on_destroy flag to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["remove_topics_on_destroy"] = remove_topics_on_destroy
        return self

    def with_topics(self, topics: Iterable[str]):
        """Set topics.

        :param topics: The topics to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["topics"] = set(topics)
        return self



from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from minos.common import (
    DataDecoder,
    DataEncoder,
    Model,
    ModelType,
    SchemaEncoder,
)


class BrokerMessage(ABC, Model):
    """Broker Message base class."""

    @property
    @abstractmethod
    def topic(self) -> str:
        """Get the topic of the message.

        :return: A ``str`` value.
        """

    @property
    @abstractmethod
    def identifier(self) -> UUID:
        """Get the identifier of the message.

        :return: An ``UUID`` instance.
        """

    @property
    def should_reply(self) -> bool:
        """

        :return:
        """
        return self.reply_topic is not None

    @abstractmethod
    def set_reply_topic(self, value: Optional[str]) -> None:
        """Set the message's reply topic.

        :param value: A ``str`` value or ``None``.
        :return: This method does not return anything.
        """

    @property
    @abstractmethod
    def reply_topic(self) -> Optional[str]:
        """Get the reply topic of the message if there is someone.

        :return: A ``str`` value or ``None``.
        """

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    @abstractmethod
    def version(cls) -> int:
        """Get the version of the message.

        :return: A strictly positive ``int`` value.
        """

    @property
    @abstractmethod
    def content(self) -> Any:
        """Get the content of the message.

        :return: Any value.
        """

    @property
    @abstractmethod
    def ok(self) -> bool:
        """Check if the message is okay or not.

        :return: ``True`` if the message is okay or ``False`` otherwise.
        """

    @property
    @abstractmethod
    def status(self) -> int:
        """Get the status of the message.

        :return: An ``int`` instance.
        """

    @property
    @abstractmethod
    def headers(self) -> dict[str, str]:
        """Get the headers of the message.

        :return: A ``dict`` instance with ``str`` keys and ``str`` values.
        """

    # noinspection PyMethodParameters
    @classmethod
    def encode_schema(cls, encoder: SchemaEncoder, target: ModelType, **kwargs) -> Any:
        """Encode schema with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded schema.
        :return: The encoded schema of the instance.
        """
        if target.classname == BrokerMessage.classname:
            return super().encode_schema(encoder, target, **kwargs)

        # noinspection PyTypeChecker
        target = ModelType.build(BrokerMessage.classname, target.type_hints | {"version": int})
        return encoder.build(target, **kwargs)

    @classmethod
    def encode_data(cls, encoder: DataEncoder, target: Any, **kwargs) -> Any:
        """Encode data with the given encoder.

        :param encoder: The encoder instance.
        :param target: An optional pre-encoded data.
        :return: The encoded data of the instance.
        """
        encoded = encoder.build(target, **kwargs)
        encoded["version"] = cls.version
        return encoded

    @classmethod
    def decode_data(cls, decoder: DataDecoder, target: Any, type_: ModelType, **kwargs) -> BrokerMessage:
        """Decode data with the given decoder.

        :param decoder: The decoder instance.
        :param target: The data to be decoded.
        :param type_: The data type.
        :return: A decoded instance.
        """
        if type_.classname != BrokerMessage.classname:
            return super().decode_data(decoder, target, type_, **kwargs)

        # The versioning bifurcation must be done here.

        from .v1 import (
            BrokerMessageV1,
        )

        name = BrokerMessageV1.classname

        # noinspection PyTypeChecker
        type_ = ModelType.build(name, {n: t for n, t in type_.type_hints.items() if n != "version"})

        return decoder.build(target, type_, **kwargs)



from .abc import (
    BrokerPublisher,
)
from .compositions import (
    InMemoryQueuedKafkaBrokerPublisher,
    PostgreSqlQueuedKafkaBrokerPublisher,
)
from .kafka import (
    KafkaBrokerPublisher,
)
from .memory import (
    InMemoryBrokerPublisher,
)
from .queued import (
    BrokerPublisherQueue,
    InMemoryBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueueQueryFactory,
    QueuedBrokerPublisher,
)



from .clients import (
    BrokerClient,
)
from .collections import (
    BrokerQueue,
    InMemoryBrokerQueue,
    PostgreSqlBrokerQueue,
)
from .dispatchers import (
    BrokerDispatcher,
    BrokerRequest,
    BrokerResponse,
    BrokerResponseException,
)
from .handlers import (
    BrokerHandler,
    BrokerHandlerService,
)
from .messages import (
    REQUEST_HEADERS_CONTEXT_VAR,
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
    BrokerMessage,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerMessageV1Status,
    BrokerMessageV1Strategy,
)
from .pools import (
    BrokerClientPool,
)
from .publishers import (
    BrokerPublisher,
    BrokerPublisherQueue,
    InMemoryBrokerPublisher,
    InMemoryBrokerPublisherQueue,
    InMemoryQueuedKafkaBrokerPublisher,
    KafkaBrokerPublisher,
    PostgreSqlBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueueQueryFactory,
    PostgreSqlQueuedKafkaBrokerPublisher,
    QueuedBrokerPublisher,
)
from .subscribers import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    InMemoryBrokerSubscriber,
    InMemoryBrokerSubscriberBuilder,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    InMemoryQueuedKafkaBrokerSubscriberBuilder,
    KafkaBrokerSubscriber,
    KafkaBrokerSubscriberBuilder,
    PostgreSqlBrokerSubscriberQueue,
    PostgreSqlBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueueQueryFactory,
    PostgreSqlQueuedKafkaBrokerSubscriberBuilder,
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)



from .abc import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)
from .kafka import (
    InMemoryQueuedKafkaBrokerSubscriberBuilder,
    KafkaBrokerSubscriber,
    KafkaBrokerSubscriberBuilder,
    PostgreSqlQueuedKafkaBrokerSubscriberBuilder,
)
from .memory import (
    InMemoryBrokerSubscriber,
    InMemoryBrokerSubscriberBuilder,
)
from .queued import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueue,
    PostgreSqlBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueueQueryFactory,
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)



from .abc import (
    BrokerPublisherQueue,
)
from .memory import (
    InMemoryBrokerPublisherQueue,
)
from .pg import (
    PostgreSqlBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueueQueryFactory,
)



from __future__ import (
    annotations,
)

import re
import socket
from abc import (
    ABC,
    abstractmethod,
)
from asyncio import (
    QueueEmpty,
)
from typing import (
    Any,
    Generic,
    TypeVar,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
)


def get_host_ip() -> str:
    """Get the host ip.

    :return: A string value.
    """
    name = get_host_name()
    return get_ip(name)


def get_host_name() -> str:
    """Get the host name.

    :return: A string value.
    """
    return re.sub(r"\.(?:local|lan)", "", socket.gethostname())


def get_ip(name: str) -> str:
    """Get the ip address.

    :param name: The name to be converted to an ip.
    :return: A string value.
    """
    return socket.gethostbyname(name)


async def consume_queue(queue, max_count: int) -> None:
    """Consume ``count`` at least ``1`` and at most ``max_count`` elements from the given queue.

    :param queue: The queue to be consumed.
    :param max_count: The max count of elements to be consumed.
    :return: This function does not return anything.
    """
    await queue.get()

    c = 1
    while c < max_count:
        c += 1
        try:
            queue.get_nowait()
        except QueueEmpty:
            break


Instance = TypeVar("Instance")


class Builder(MinosSetup, ABC, Generic[Instance]):  # FIXME: This class should be part of `minos.common` or similar.
    """Builder class."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kwargs = dict()

    def copy(self: type[B]) -> B:
        """Get a copy of the instance.

        :return: A ``BrokerSubscriberBuilder`` instance.
        """
        return self.new().with_kwargs(self.kwargs)

    @classmethod
    def new(cls: type[B]) -> B:
        """Get a new instance.

        :return: A ``BrokerSubscriberBuilder`` instance.
        """
        return cls()

    def with_kwargs(self: B, kwargs: dict[str, Any]) -> B:
        """Set kwargs.

        :param kwargs: The kwargs to be set.
        :return: This method return the builder instance.
        """
        self.kwargs |= kwargs
        return self

    def with_config(self: B, config: MinosConfig) -> B:
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        return self

    @abstractmethod
    def build(self) -> Instance:
        """Build the instance.

        :return: A ``BrokerSubscriber`` instance.
        """


B = TypeVar("B", bound=Builder)



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    TimeoutError,
    wait_for,
)
from collections.abc import (
    Iterable,
)
from contextlib import (
    suppress,
)
from typing import (
    Optional,
)

from aiokafka import (
    AIOKafkaConsumer,
)
from cached_property import (
    cached_property,
)
from kafka import (
    KafkaAdminClient,
)
from kafka.admin import (
    NewTopic,
)
from kafka.errors import (
    TopicAlreadyExistsError,
)

from minos.common import (
    MinosConfig,
)

from ..messages import (
    BrokerMessage,
)
from .abc import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)
from .queued import (
    InMemoryBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueueBuilder,
    QueuedBrokerSubscriberBuilder,
)

logger = logging.getLogger(__name__)


class KafkaBrokerSubscriber(BrokerSubscriber):
    """Kafka Broker Subscriber class."""

    def __init__(
        self,
        topics: Iterable[str],
        broker_host: str,
        broker_port: int,
        group_id: Optional[str] = None,
        remove_topics_on_destroy: bool = False,
        **kwargs,
    ):
        super().__init__(topics, **kwargs)
        self.broker_host = broker_host
        self.broker_port = broker_port
        self.group_id = group_id

        self.remove_topics_on_destroy = remove_topics_on_destroy

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> KafkaBrokerSubscriber:
        # noinspection PyTypeChecker
        return KafkaBrokerSubscriberBuilder.new().with_config(config).with_kwargs(kwargs).build()

    async def _setup(self) -> None:
        await super()._setup()
        self._create_topics()
        await self.client.start()

    async def _destroy(self) -> None:
        with suppress(TimeoutError):
            await wait_for(self.client.stop(), 0.5)
        self._delete_topics()
        self.admin_client.close()
        await super()._destroy()

    def _create_topics(self) -> None:
        logger.info(f"Creating {self.topics!r} topics...")

        new_topics = list()
        for topic in self.topics:
            new_topics.append(NewTopic(name=topic, num_partitions=1, replication_factor=1))

        with suppress(TopicAlreadyExistsError):
            self.admin_client.create_topics(new_topics)

    def _delete_topics(self) -> None:
        if not self.remove_topics_on_destroy:
            return

        logger.info(f"Deleting {self.topics!r} topics...")
        self.admin_client.delete_topics(list(self.topics))

    @cached_property
    def admin_client(self):
        """Get the kafka admin client.

        :return: An ``KafkaAdminClient`` instance.
        """
        return KafkaAdminClient(bootstrap_servers=f"{self.broker_host}:{self.broker_port}")

    async def _receive(self) -> BrokerMessage:
        record = await self.client.getone()
        bytes_ = record.value
        message = BrokerMessage.from_avro_bytes(bytes_)
        return message

    @cached_property
    def client(self) -> AIOKafkaConsumer:
        """Get the kafka consumer client.

        :return: An ``AIOKafkaConsumer`` instance.
        """
        return AIOKafkaConsumer(
            *self.topics,
            bootstrap_servers=f"{self.broker_host}:{self.broker_port}",
            group_id=self.group_id,
            auto_offset_reset="earliest",
        )


class KafkaBrokerSubscriberBuilder(BrokerSubscriberBuilder):
    """Kafka Broker Subscriber Builder class."""

    def with_config(self, config: MinosConfig) -> BrokerSubscriberBuilder:
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self.kwargs |= {
            "group_id": config.service.name,
            "broker_host": config.broker.host,
            "broker_port": config.broker.port,
        }
        return self

    def build(self) -> BrokerSubscriber:
        """Build the instance.

        :return: A ``KafkaBrokerSubscriber`` instance.
        """
        return KafkaBrokerSubscriber(**self.kwargs)


class PostgreSqlQueuedKafkaBrokerSubscriberBuilder(QueuedBrokerSubscriberBuilder):
    """PostgreSql Queued Kafka Broker Subscriber Builder class."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            impl_builder=KafkaBrokerSubscriberBuilder.new(),
            queue_builder=PostgreSqlBrokerSubscriberQueueBuilder.new(),
            **kwargs,
        )


class InMemoryQueuedKafkaBrokerSubscriberBuilder(QueuedBrokerSubscriberBuilder):
    """In Memory Queued Kafka Broker Subscriber Builder class."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            *args,
            impl_builder=KafkaBrokerSubscriberBuilder.new(),
            queue_builder=InMemoryBrokerSubscriberQueueBuilder.new(),
            **kwargs,
        )



from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    Optional,
)

from aiopg import (
    Cursor,
)
from psycopg2.sql import (
    SQL,
    Identifier,
)

from minos.common import (
    MinosConfig,
)

from ....collections import (
    PostgreSqlBrokerQueue,
    PostgreSqlBrokerQueueQueryFactory,
)
from ....messages import (
    BrokerMessage,
)
from .abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)

logger = logging.getLogger(__name__)


class PostgreSqlBrokerSubscriberQueue(PostgreSqlBrokerQueue, BrokerSubscriberQueue):
    """PostgreSql Broker Subscriber Queue class."""

    def __init__(
        self, topics: set[str], *args, query_factory: Optional[PostgreSqlBrokerQueueQueryFactory] = None, **kwargs
    ):
        if query_factory is None:
            query_factory = PostgreSqlBrokerSubscriberQueueQueryFactory()
        super().__init__(topics, *args, query_factory=query_factory, **kwargs)

    async def _notify_enqueued(self, message: BrokerMessage) -> None:
        await self.submit_query(self._query_factory.build_notify().format(Identifier(message.topic)))

    async def _listen_entries(self, cursor: Cursor) -> None:
        for topic in self.topics:
            await cursor.execute(self._query_factory.build_listen().format(Identifier(topic)))

    async def _unlisten_entries(self, cursor: Cursor) -> None:
        if not cursor.closed:
            for topic in self.topics:
                await cursor.execute(self._query_factory.build_unlisten().format(Identifier(topic)))

    async def _get_count(self, cursor: Cursor) -> int:
        # noinspection PyTypeChecker
        await cursor.execute(self._query_factory.build_count_not_processed(), (self._retry, tuple(self.topics)))
        count = (await cursor.fetchone())[0]
        return count

    async def _dequeue_rows(self, cursor: Cursor) -> list[Any]:
        # noinspection PyTypeChecker
        await cursor.execute(
            self._query_factory.build_select_not_processed(), (self._retry, tuple(self.topics), self._records)
        )
        return await cursor.fetchall()


class PostgreSqlBrokerSubscriberQueueQueryFactory(PostgreSqlBrokerQueueQueryFactory):
    """PostgreSql Broker Subscriber Queue Query Factory class."""

    def build_table_name(self) -> str:
        """Get the table name.

        :return: A ``str`` value.
        """
        return "broker_subscriber_queue"

    def build_notify(self) -> SQL:
        """Build the "notify" query.

        :return: A ``SQL`` instance.
        """
        return SQL("NOTIFY {}")

    def build_listen(self) -> SQL:
        """Build the "listen" query.

        :return: A ``SQL`` instance.
        """
        return SQL("LISTEN {}")

    def build_unlisten(self) -> SQL:
        """Build the "unlisten" query.

        :return: A ``SQL`` instance.
        """
        return SQL("UNLISTEN {}")

    def build_count_not_processed(self) -> SQL:
        """Build the "count not processed" query.

        :return:
        """
        return SQL(
            f"SELECT COUNT(*) FROM (SELECT id FROM {self.build_table_name()} "
            "WHERE NOT processing AND retry < %s AND topic IN %s FOR UPDATE SKIP LOCKED) s"
        )

    def build_select_not_processed(self) -> SQL:
        """Build the "select not processed" query.

        :return: A ``SQL`` instance.
        """
        return SQL(
            "SELECT id, data "
            f"FROM {self.build_table_name()} "
            "WHERE NOT processing AND retry < %s AND topic IN %s "
            "ORDER BY created_at "
            "LIMIT %s "
            "FOR UPDATE SKIP LOCKED"
        )


class PostgreSqlBrokerSubscriberQueueBuilder(BrokerSubscriberQueueBuilder):
    """PostgreSql Broker Subscriber Queue Builder class."""

    def with_config(self, config: MinosConfig):
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        # noinspection PyProtectedMember
        self.kwargs |= config.broker.queue._asdict()
        return super().with_config(config)

    def build(self) -> PostgreSqlBrokerSubscriberQueue:
        """Build the instance.

        :return: A ``BrokerSubscriberQueue`` instance.
        """
        return PostgreSqlBrokerSubscriberQueue(**self.kwargs)



from __future__ import (
    annotations,
)

import logging

from ....collections import (
    InMemoryBrokerQueue,
)
from .abc import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)

logger = logging.getLogger(__name__)


class InMemoryBrokerSubscriberQueue(InMemoryBrokerQueue, BrokerSubscriberQueue):
    """In Memory Broker Subscriber Queue class."""


class InMemoryBrokerSubscriberQueueBuilder(BrokerSubscriberQueueBuilder):
    """In Memory Broker Subscriber Queue Builder class."""

    def build(self) -> BrokerSubscriberQueue:
        """Build the instance.

        :return: An ``InMemoryBrokerSubscriberQueue`` instance.
        """
        return InMemoryBrokerSubscriberQueue(**self.kwargs)



from .impl import (
    QueuedBrokerSubscriber,
    QueuedBrokerSubscriberBuilder,
)
from .queues import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
    InMemoryBrokerSubscriberQueue,
    InMemoryBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueue,
    PostgreSqlBrokerSubscriberQueueBuilder,
    PostgreSqlBrokerSubscriberQueueQueryFactory,
)



from __future__ import (
    annotations,
)

import logging
from asyncio import (
    TimeoutError,
    wait_for,
)
from collections.abc import (
    AsyncIterator,
)
from typing import (
    Optional,
)
from uuid import (
    uuid4,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
    NotProvidedException,
)

from ..exceptions import (
    MinosHandlerNotFoundEnoughEntriesException,
)
from .messages import (
    BrokerMessage,
)
from .publishers import (
    BrokerPublisher,
)
from .subscribers import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)

logger = logging.getLogger(__name__)


class BrokerClient(MinosSetup):
    """Broker Client class."""

    def __init__(self, topic: str, publisher: BrokerPublisher, subscriber: BrokerSubscriber, **kwargs):
        super().__init__(**kwargs)

        self.topic = topic
        self.publisher = publisher
        self.subscriber = subscriber

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerClient:
        if "topic" not in kwargs:
            kwargs["topic"] = str(uuid4()).replace("-", "")

        kwargs["publisher"] = cls._get_publisher(**kwargs)

        kwargs["subscriber"] = cls._get_subscriber(config, **kwargs)
        # noinspection PyProtectedMember
        return cls(**kwargs)

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_publisher(
        publisher: Optional[BrokerPublisher] = None,
        broker_publisher: BrokerPublisher = Provide["broker_publisher"],
        **kwargs,
    ) -> BrokerPublisher:
        if publisher is None:
            publisher = broker_publisher
        if publisher is None or isinstance(publisher, Provide):
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        return publisher

    @staticmethod
    @inject
    def _get_subscriber(
        config: MinosConfig,
        topic: str,
        subscriber: Optional[BrokerSubscriber] = None,
        broker_subscriber: Optional[BrokerSubscriber] = Provide["broker_subscriber"],
        subscriber_builder: Optional[BrokerSubscriberBuilder] = None,
        broker_subscriber_builder: Optional[BrokerSubscriberBuilder] = Provide["broker_subscriber_builder"],
        **kwargs,
    ) -> BrokerSubscriber:
        if not isinstance(subscriber, BrokerSubscriber):
            subscriber = broker_subscriber

        if not isinstance(subscriber, BrokerSubscriber):
            if not isinstance(subscriber_builder, BrokerSubscriberBuilder):
                subscriber_builder = broker_subscriber_builder

            if isinstance(subscriber_builder, BrokerSubscriberBuilder):
                subscriber = (
                    subscriber_builder.copy()
                    .with_config(config)
                    .with_topics({topic})
                    .with_group_id(None)
                    .with_remove_topics_on_destroy(True)
                    .with_kwargs(kwargs)
                    .build()
                )

        if not isinstance(subscriber, BrokerSubscriber):
            raise NotProvidedException(f"A {BrokerSubscriber!r} or {BrokerSubscriberBuilder!r} must be provided.")

        return subscriber

    async def _setup(self) -> None:
        await super()._setup()
        await self.subscriber.setup()

    async def _destroy(self) -> None:
        await self.subscriber.destroy()
        await super()._destroy()

    # noinspection PyUnusedLocal
    async def send(self, message: BrokerMessage) -> None:
        """Send a ``BrokerMessage``.

        :param message: The message to be sent.
        :return: This method does not return anything.
        """
        message.set_reply_topic(self.topic)
        await self.publisher.send(message)

    async def receive(self, *args, **kwargs) -> BrokerMessage:
        """Get one handler entry from the given topics.

        :param args: Additional positional parameters to be passed to receive_many.
        :param kwargs: Additional named parameters to be passed to receive_many.
        :return: A ``HandlerEntry`` instance.
        """
        return await self.receive_many(*args, **(kwargs | {"count": 1})).__anext__()

    async def receive_many(self, count: int, timeout: float = 60, **kwargs) -> AsyncIterator[BrokerMessage]:
        """Get multiple handler entries from the given topics.

        :param timeout: Maximum time in seconds to wait for messages.
        :param count: Number of entries to be collected.
        :return: A list of ``HandlerEntry`` instances.
        """
        try:
            messages = await wait_for(self._get_many(count, **kwargs), timeout=timeout)
        except TimeoutError:
            raise MinosHandlerNotFoundEnoughEntriesException(
                f"Timeout exceeded while trying to fetch {count!r} entries from {self.topic!r}."
            )

        for message in messages:
            logger.info(f"Dispatching '{message!s}'...")
            yield message

    async def _get_many(self, count, *args, **kwargs) -> list[BrokerMessage]:
        result = list()
        async for message in self.subscriber:
            result.append(message)

            if len(result) == count:
                break

        return result



from __future__ import (
    annotations,
)

import logging
from collections.abc import (
    Awaitable,
    Callable,
)
from functools import (
    wraps,
)
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Optional,
    Union,
)

from dependency_injector.wiring import (
    Provide,
    inject,
)

from minos.common import (
    MinosConfig,
    MinosSetup,
    NotProvidedException,
)

from ...decorators import (
    EnrouteBuilder,
)
from ...exceptions import (
    MinosActionNotFoundException,
)
from ...requests import (
    REQUEST_USER_CONTEXT_VAR,
    Request,
    Response,
    ResponseException,
)
from ..messages import (
    REQUEST_HEADERS_CONTEXT_VAR,
    BrokerMessage,
    BrokerMessageV1,
    BrokerMessageV1Payload,
    BrokerMessageV1Status,
)
from ..publishers import (
    BrokerPublisher,
)
from .requests import (
    BrokerRequest,
    BrokerResponse,
)

logger = logging.getLogger(__name__)


class BrokerDispatcher(MinosSetup):
    """Broker Dispatcher class."""

    def __init__(self, actions: dict[str, Optional[Callable]], publisher: BrokerPublisher, **kwargs):
        super().__init__(**kwargs)
        self._actions = actions
        self._publisher = publisher

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerDispatcher:
        kwargs["actions"] = cls._get_actions(config, **kwargs)
        kwargs["publisher"] = cls._get_publisher(**kwargs)
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    @staticmethod
    def _get_actions(
        config: MinosConfig, handlers: dict[str, Optional[Callable]] = None, **kwargs
    ) -> dict[str, Callable[[BrokerRequest], Awaitable[Optional[BrokerResponse]]]]:
        if handlers is None:
            builder = EnrouteBuilder(*config.services, middleware=config.middleware)
            decorators = builder.get_broker_command_query_event(config=config, **kwargs)
            handlers = {decorator.topic: fn for decorator, fn in decorators.items()}
        return handlers

    # noinspection PyUnusedLocal
    @staticmethod
    @inject
    def _get_publisher(
        publisher: Optional[BrokerPublisher] = None,
        broker_publisher: BrokerPublisher = Provide["broker_publisher"],
        **kwargs,
    ) -> BrokerPublisher:
        if publisher is None:
            publisher = broker_publisher
        if publisher is None or isinstance(publisher, Provide):
            raise NotProvidedException(f"A {BrokerPublisher!r} object must be provided.")
        return publisher

    @property
    def publisher(self) -> BrokerPublisher:
        """Get the publisher instance.

        :return: A ``BrokerPublisher`` instance.
        """
        return self._publisher

    @property
    def actions(self) -> dict[str, Optional[Callable]]:
        """Actions getter.

        :return: A dictionary in which the keys are topics and the values are the handler.
        """
        return self._actions

    async def dispatch(self, message: BrokerMessage) -> None:
        """Dispatch an entry.

        :param message: The entry to be dispatched.
        :return: This method does not return anything.
        """
        action = self.get_action(message.topic)
        fn = self.get_callback(action)

        data, status, headers = await fn(message)

        if message.should_reply:
            reply = BrokerMessageV1(
                topic=message.reply_topic,
                payload=BrokerMessageV1Payload(content=data, status=status, headers=headers),
                identifier=message.identifier,
            )
            await self.publisher.send(reply)

    @staticmethod
    def get_callback(
        fn: Callable[[BrokerRequest], Union[Optional[BrokerResponse], Awaitable[Optional[BrokerResponse]]]]
    ) -> Callable[[BrokerMessage], Awaitable[tuple[Any, int, dict[str, str]]]]:
        """Get the handler function to be used by the Broker Handler.

        :param fn: The action function.
        :return: A wrapper function around the given one that is compatible with the Broker Handler API.
        """

        @wraps(fn)
        async def _wrapper(raw: BrokerMessage) -> tuple[Any, int, dict[str, str]]:
            logger.info(f"Dispatching '{raw!s}'...")

            request = BrokerRequest(raw)
            user_token = REQUEST_USER_CONTEXT_VAR.set(request.user)
            headers_token = REQUEST_HEADERS_CONTEXT_VAR.set(request.headers)

            try:
                response = fn(request)
                if isawaitable(response):
                    response = await response
                if isinstance(response, Response):
                    response = await response.content()
                return response, BrokerMessageV1Status.SUCCESS, REQUEST_HEADERS_CONTEXT_VAR.get()
            except ResponseException as exc:
                logger.warning(f"Raised an application exception: {exc!s}")
                return repr(exc), BrokerMessageV1Status.ERROR, REQUEST_HEADERS_CONTEXT_VAR.get()
            except Exception as exc:
                logger.exception(f"Raised a system exception: {exc!r}")
                return repr(exc), BrokerMessageV1Status.SYSTEM_ERROR, REQUEST_HEADERS_CONTEXT_VAR.get()
            finally:
                REQUEST_USER_CONTEXT_VAR.reset(user_token)
                REQUEST_HEADERS_CONTEXT_VAR.reset(headers_token)

        return _wrapper

    def get_action(self, topic: str) -> Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]:
        """Get the handling function to be called.

        :param topic: The name of the topic that matches the action.
        :return: A ``Callable`` instance.
        """
        if topic not in self._actions:
            raise MinosActionNotFoundException(
                f"topic {topic} have no controller/action configured, " f"please review th configuration file"
            )

        action = self._actions[topic]

        logger.debug(f"Loaded {action!r} action!")
        return action



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from asyncio import (
    CancelledError,
    PriorityQueue,
    QueueEmpty,
    TimeoutError,
    create_task,
    wait_for,
)
from contextlib import (
    suppress,
)
from typing import (
    Any,
    NoReturn,
    Optional,
)

from aiopg import (
    Cursor,
)
from cached_property import (
    cached_property,
)
from psycopg2.sql import (
    SQL,
)

from minos.common import (
    MinosConfig,
    PostgreSqlMinosDatabase,
)

from ....utils import (
    consume_queue,
)
from ...messages import (
    BrokerMessage,
)
from .abc import (
    BrokerQueue,
)

logger = logging.getLogger(__name__)


class PostgreSqlBrokerQueue(BrokerQueue, PostgreSqlMinosDatabase):
    """PostgreSql Broker Queue class."""

    _queue: PriorityQueue[_Entry]

    def __init__(self, *args, query_factory: PostgreSqlBrokerQueueQueryFactory, retry: int, records: int, **kwargs):
        super().__init__(*args, **kwargs)

        self._query_factory = query_factory
        self._retry = retry
        self._records = records

        self._queue = PriorityQueue(maxsize=records)

        self._run_task = None

    @property
    def query_factory(self) -> PostgreSqlBrokerQueueQueryFactory:
        """Get the query factory.

        :return: A ``PostgreSqlBrokerQueueQueryFactory`` instance.
        """
        return self._query_factory

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> PostgreSqlBrokerQueue:
        # noinspection PyProtectedMember
        return cls(**config.broker.queue._asdict(), **kwargs)

    async def _setup(self) -> None:
        await super()._setup()
        await self._create_table()
        await self._start_run()

    async def _destroy(self) -> None:
        await self._stop_run()
        await self._flush_queue()
        await super()._destroy()

    async def _create_table(self) -> None:
        await self.submit_query(self._query_factory.build_create_table(), lock=self._query_factory.build_table_name())

    async def _start_run(self) -> None:
        if self._run_task is None:
            self._run_task = create_task(self._run())

    async def _stop_run(self) -> None:
        if self._run_task is not None:
            task = self._run_task
            self._run_task = None

            task.cancel()
            with suppress(TimeoutError, CancelledError):
                await wait_for(task, 0.5)

    async def _flush_queue(self):
        while True:
            try:
                entry = self._queue.get_nowait()
            except QueueEmpty:
                break
            await self.submit_query(self._query_factory.build_update_not_processed(), (entry.id_,))
            self._queue.task_done()

    async def _enqueue(self, message: BrokerMessage) -> None:
        await self.submit_query_and_fetchone(self._query_factory.build_insert(), (message.topic, message.avro_bytes))
        await self._notify_enqueued(message)

    async def _notify_enqueued(self, message: BrokerMessage) -> None:
        await self.submit_query(self._query_factory.build_notify())

    async def _dequeue(self) -> BrokerMessage:
        while True:
            entry = await self._queue.get()
            try:
                # noinspection PyBroadException
                try:
                    message = entry.data
                except Exception as exc:
                    logger.warning(
                        f"There was a problem while trying to deserialize the entry with {entry.id_!r} id: {exc}"
                    )
                    await self.submit_query(self._query_factory.build_update_not_processed(), (entry.id_,))
                    continue

                await self.submit_query(self._query_factory.build_delete_processed(), (entry.id_,))
                return message
            finally:
                self._queue.task_done()

    async def _run(self, max_wait: Optional[float] = 60.0) -> NoReturn:
        async with self.cursor() as cursor:
            await self._listen_entries(cursor)
            try:
                while self._run_task is not None:
                    await self._wait_for_entries(cursor, max_wait)
                    await self._dequeue_batch(cursor)
            finally:
                await self._unlisten_entries(cursor)

    async def _listen_entries(self, cursor: Cursor) -> None:
        # noinspection PyTypeChecker
        await cursor.execute(self._query_factory.build_listen())

    async def _unlisten_entries(self, cursor: Cursor) -> None:
        if not cursor.closed:
            # noinspection PyTypeChecker
            await cursor.execute(self._query_factory.build_unlisten())

    async def _wait_for_entries(self, cursor: Cursor, max_wait: Optional[float]) -> None:
        while True:
            if await self._get_count(cursor):
                return

            with suppress(TimeoutError):
                return await wait_for(consume_queue(cursor.connection.notifies, self._records), max_wait)

    async def _get_count(self, cursor: Cursor) -> int:
        # noinspection PyTypeChecker
        await cursor.execute(self._query_factory.build_count_not_processed(), (self._retry,))
        count = (await cursor.fetchone())[0]
        return count

    async def _dequeue_batch(self, cursor: Cursor) -> None:
        async with cursor.begin():
            rows = await self._dequeue_rows(cursor)

            if not len(rows):
                return

            entries = [_Entry(*row) for row in rows]

            # noinspection PyTypeChecker
            await cursor.execute(self._query_factory.build_mark_processing(), (tuple(entry.id_ for entry in entries),))

            for entry in entries:
                await self._queue.put(entry)

    async def _dequeue_rows(self, cursor: Cursor) -> list[Any]:
        # noinspection PyTypeChecker
        await cursor.execute(self._query_factory.build_select_not_processed(), (self._retry, self._records))
        return await cursor.fetchall()


class PostgreSqlBrokerQueueQueryFactory(ABC):
    """PostgreSql Broker Queue Query Factory class."""

    @abstractmethod
    def build_table_name(self) -> str:
        """Get the table name.

        :return: A ``str`` value.
        """
        raise NotImplementedError

    def build_create_table(self) -> SQL:
        """Build the "create table" query.

        :return: A ``SQL`` instance.
        """
        return SQL(
            f"CREATE TABLE IF NOT EXISTS {self.build_table_name()} ("
            "id BIGSERIAL NOT NULL PRIMARY KEY, "
            "topic VARCHAR(255) NOT NULL, "
            "data BYTEA NOT NULL, "
            "retry INTEGER NOT NULL DEFAULT 0, "
            "processing BOOL NOT NULL DEFAULT FALSE, "
            "created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(), "
            "updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW())"
        )

    def build_update_not_processed(self) -> SQL:
        """Build the "update not processed" query.

        :return: A ``SQL`` instance.
        """
        return SQL(
            f"UPDATE {self.build_table_name()} "
            "SET processing = FALSE, retry = retry + 1, updated_at = NOW() WHERE id = %s"
        )

    def build_delete_processed(self) -> SQL:
        """Build the "delete processed" query.

        :return: A ``SQL`` instance.
        """
        return SQL(f"DELETE FROM {self.build_table_name()} WHERE id = %s")

    def build_mark_processing(self) -> SQL:
        """

        :return: A ``SQL`` instance.
        """
        return SQL(f"UPDATE {self.build_table_name()} SET processing = TRUE WHERE id IN %s")

    def build_notify(self) -> SQL:
        """Build the "notify" query.

        :return: A ``SQL`` instance.
        """
        return SQL(f"NOTIFY {self.build_table_name()}")

    def build_listen(self) -> SQL:
        """Build the "listen" query.

        :return: A ``SQL`` instance.
        """
        return SQL(f"LISTEN {self.build_table_name()}")

    def build_unlisten(self) -> SQL:
        """Build the "unlisten" query.

        :return: A ``SQL`` instance.
        """
        return SQL(f"UNLISTEN {self.build_table_name()}")

    def build_count_not_processed(self) -> SQL:
        """Build the "count not processed" query.

        :return:
        """
        return SQL(
            f"SELECT COUNT(*) FROM (SELECT id FROM {self.build_table_name()} "
            "WHERE NOT processing AND retry < %s FOR UPDATE SKIP LOCKED) s"
        )

    def build_insert(self) -> SQL:
        """Build the "insert" query.

        :return: A ``SQL`` instance.
        """
        return SQL(f"INSERT INTO {self.build_table_name()} (topic, data) VALUES (%s, %s) RETURNING id")

    def build_select_not_processed(self) -> SQL:
        """Build the "select not processed" query.

        :return: A ``SQL`` instance.
        """
        return SQL(
            "SELECT id, data "
            f"FROM {self.build_table_name()} "
            "WHERE NOT processing AND retry < %s "
            "ORDER BY created_at "
            "LIMIT %s "
            "FOR UPDATE "
            "SKIP LOCKED"
        )


class _Entry:
    def __init__(self, id_: int, data_bytes: bytes):
        self.id_ = id_
        self.data_bytes = data_bytes

    @cached_property
    def data(self) -> BrokerMessage:
        """Get the data.

        :return: A ``Model`` inherited instance.
        """
        return BrokerMessage.from_avro_bytes(self.data_bytes)

    def __lt__(self, other: Any) -> bool:
        # noinspection PyBroadException
        try:
            return isinstance(other, type(self)) and self.data < other.data
        except Exception:
            return False



from __future__ import (
    annotations,
)

import logging
from abc import (
    ABC,
    abstractmethod,
)
from collections.abc import (
    AsyncIterator,
)

from minos.common import (
    MinosSetup,
)

from ...messages import (
    BrokerMessage,
)

logger = logging.getLogger(__name__)


class BrokerQueue(ABC, MinosSetup):
    """Broker Queue class."""

    async def enqueue(self, message: BrokerMessage) -> None:
        """Enqueue method."""
        logger.info(f"Enqueuing {message!r} message...")
        await self._enqueue(message)

    @abstractmethod
    async def _enqueue(self, message: BrokerMessage) -> None:
        raise NotImplementedError

    def __aiter__(self) -> AsyncIterator[BrokerMessage]:
        return self

    async def __anext__(self) -> BrokerMessage:
        if self.already_destroyed:
            raise StopAsyncIteration
        return await self.dequeue()

    async def dequeue(self) -> BrokerMessage:
        """Dequeue method."""
        message = await self._dequeue()
        logger.info(f"Dequeuing {message!r} message...")
        return message

    @abstractmethod
    async def _dequeue(self) -> BrokerMessage:
        raise NotImplementedError



from asyncio import (
    CancelledError,
    TimeoutError,
    create_task,
    wait_for,
)
from contextlib import (
    suppress,
)
from typing import (
    NoReturn,
)

from ...messages import (
    BrokerMessage,
)
from ..abc import (
    BrokerPublisher,
)
from .queues import (
    BrokerPublisherQueue,
)


class QueuedBrokerPublisher(BrokerPublisher):
    """Queued Broker Publisher class."""

    impl: BrokerPublisher
    queue: BrokerPublisherQueue

    def __init__(self, impl: BrokerPublisher, queue: BrokerPublisherQueue, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.impl = impl
        self.queue = queue

        self._run_task = None

    async def _setup(self) -> None:
        await super()._setup()
        await self.queue.setup()
        await self.impl.setup()
        await self._start_task()

    async def _destroy(self) -> None:
        await self._stop_task()
        await self.impl.destroy()
        await self.queue.destroy()
        await super()._destroy()

    async def _start_task(self):
        if self._run_task is None:
            self._run_task = create_task(self._run())

    async def _stop_task(self):
        if self._run_task is not None:
            self._run_task.cancel()
            with suppress(TimeoutError, CancelledError):
                await wait_for(self._run_task, 0.5)
            self._run_task = None

    async def _run(self) -> NoReturn:
        async for message in self.queue:
            await self.impl.send(message)

    async def _send(self, message: BrokerMessage) -> None:
        await self.queue.enqueue(message)



from __future__ import (
    annotations,
)

from asyncio import (
    iscoroutinefunction,
)
from collections.abc import (
    Awaitable,
    Callable,
    Iterable,
)
from functools import (
    partial,
    wraps,
)
from inspect import (
    isawaitable,
)
from typing import (
    TYPE_CHECKING,
    Optional,
    Protocol,
    Type,
    Union,
    runtime_checkable,
)

from cached_property import (
    cached_property,
)

from ...exceptions import (
    MinosMultipleEnrouteDecoratorKindsException,
    NotSatisfiedCheckerException,
)
from ...requests import (
    Request,
    Response,
    ResponseException,
)
from .checkers import (
    CheckerMeta,
)

if TYPE_CHECKING:
    from ..definitions import (
        CheckDecorator,
        EnrouteDecorator,
    )

Handler = Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]


@runtime_checkable
class HandlerWrapper(Protocol):
    """Handler Wrapper class."""

    meta: HandlerMeta
    check: Type[CheckDecorator]
    __call__: Handler


class HandlerMeta:
    """Handler Meta class."""

    func: Handler
    decorators: set[EnrouteDecorator]
    checkers: set[CheckerMeta]

    def __init__(
        self,
        func: Handler,
        decorators: Optional[set[EnrouteDecorator]] = None,
        checkers: Optional[set[CheckerMeta]] = None,
    ):
        if decorators is None:
            decorators = set()
        if checkers is None:
            checkers = set()
        self.func = func
        self.decorators = decorators
        self.checkers = checkers

    @property
    def wrapper(self) -> HandlerWrapper:
        """Get the ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """
        if iscoroutinefunction(self.func):
            return self.async_wrapper
        else:
            return self.sync_wrapper

    @cached_property
    def async_wrapper(self) -> HandlerWrapper:
        """Get the async ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """

        @wraps(self.func)
        async def _wrapper(*args, **kwargs) -> Optional[Response]:
            try:
                await CheckerMeta.run_async(self.checkers, *args, **kwargs)
            except NotSatisfiedCheckerException as exc:
                raise ResponseException(f"There was an exception during check step: {exc}")

            response = self.func(*args, **kwargs)
            if isawaitable(response):
                response = await response
            return response

        _wrapper.meta = self
        _wrapper.check = self.check
        _wrapper.__decorators__ = self.decorators  # FIXME: This attribute should be removed in future versions.

        return _wrapper

    @cached_property
    def sync_wrapper(self) -> HandlerWrapper:
        """Get the sync ``HandlerWrapper`` instance.

        :return: A ``HandlerWrapper`` instance.
        """

        if iscoroutinefunction(self.func):
            raise ValueError(f"{self.func!r} cannot be awaitable.")

        @wraps(self.func)
        def _wrapper(*args, **kwargs) -> Optional[Response]:
            try:
                CheckerMeta.run_sync(self.checkers, *args, **kwargs)
            except NotSatisfiedCheckerException as exc:
                raise ResponseException(f"There was an exception during check step: {exc}")

            return self.func(*args, **kwargs)

        _wrapper.meta = self
        _wrapper.check = self.check
        _wrapper.__decorators__ = self.decorators  # FIXME: This attribute should be removed in future versions.

        return _wrapper

    def add_decorator(self, decorator: EnrouteDecorator) -> None:
        """Add a new decorator to the ``decorators`` set.

        :param decorator: The decorator to be added.
        :return: This method does not return anything.
        """
        another = next(iter(self.decorators), None)
        if another is not None and another.KIND != decorator.KIND:
            raise MinosMultipleEnrouteDecoratorKindsException(
                f"There are multiple kinds but only one is allowed: {(another.KIND, decorator.KIND)}"
            )
        self.decorators.add(decorator)

    @cached_property
    def check(self) -> Type[CheckDecorator]:
        """Get the check decorator.

        :return: A ``CheckDecorator`` type.
        """
        from ..definitions import (
            CheckDecorator,
        )

        # noinspection PyTypeChecker
        return partial(CheckDecorator, handler_meta=self)

    def __repr__(self):
        args = ", ".join(map(repr, self))
        return f"{type(self).__name__}({args})"

    def __eq__(self, other: CheckerMeta) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(self.func)

    def __iter__(self) -> Iterable:
        yield from (
            self.func,
            self.decorators,
            self.checkers,
        )



import logging
from abc import (
    ABC,
    abstractmethod,
)

from minos.common import (
    MinosSetup,
)

from ..messages import (
    BrokerMessage,
)

logger = logging.getLogger(__name__)


class BrokerPublisher(ABC, MinosSetup):
    """Broker Publisher class."""

    async def send(self, message: BrokerMessage) -> None:
        """Send a message.

        :param message: The message to be sent.
        :return: This method does not return anything.
        """
        logger.info(f"Sending {message!r} message...")
        await self._send(message)

    @abstractmethod
    async def _send(self, message: BrokerMessage) -> None:
        raise NotImplementedError



from .impl import (
    QueuedBrokerPublisher,
)
from .queues import (
    BrokerPublisherQueue,
    InMemoryBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueue,
    PostgreSqlBrokerPublisherQueueQueryFactory,
)



from contextvars import (
    ContextVar,
)
from typing import (
    Final,
    Optional,
)

REQUEST_REPLY_TOPIC_CONTEXT_VAR: Final[ContextVar[Optional[str]]] = ContextVar("reply_topic", default=None)
REQUEST_HEADERS_CONTEXT_VAR: Final[ContextVar[Optional[dict[str, str]]]] = ContextVar("headers", default=None)



from __future__ import (
    annotations,
)

import warnings
from collections import (
    defaultdict,
)
from collections.abc import (
    Callable,
    Iterable,
)
from itertools import (
    chain,
)
from typing import (
    Any,
    Awaitable,
    Optional,
    Union,
)
from urllib.parse import (
    parse_qsl,
    urlencode,
)
from uuid import (
    UUID,
)

from aiohttp.web import Request as AioHttpRequest
from cached_property import (
    cached_property,
)
from orjson import (
    orjson,
)

from minos.common import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
    MinosAvroProtocol,
    TypeHintBuilder,
    import_module,
)

from ..exceptions import (
    NotHasParamsException,
)
from ..requests import (
    Request,
    Response,
    ResponseException,
)


class RestRequest(Request):
    """Rest Request class."""

    __slots__ = "raw"

    def __init__(self, raw: AioHttpRequest, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.raw = raw

    @property
    def raw_request(self) -> AioHttpRequest:
        """Get the raw request within the instance.

        :return: An ``aiohttp.web.Request`` instance.
        """
        warnings.warn(
            f"'{RestRequest.__name__}s.raw_request' is deprecated in favor of '{RestRequest.__name__}.raw'.",
            DeprecationWarning,
        )
        return self.raw

    def __eq__(self, other: RestRequest) -> bool:
        return type(self) == type(other) and self.raw == other.raw

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.raw!r})"

    @cached_property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        if "user" not in self.headers:
            return None
        return UUID(self.headers["user"])

    @property
    def headers(self) -> dict[str, str]:
        """Get the headers of the request.

        :return: A dictionary in which keys are ``str`` instances and values are ``str`` instances.
        """
        # noinspection PyTypeChecker
        return self.raw.headers

    @property
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        return self.raw.body_exists

    async def _content(self, type_: Optional[Union[type, str]] = None, **kwargs) -> Any:
        if "model_type" in kwargs:
            warnings.warn("The 'model_type' argument is deprecated. Use 'type_' instead", DeprecationWarning)
            if type_ is None:
                type_ = kwargs["model_type"]

        data = await self._content_parser()
        return self._build(data, type_)

    @cached_property
    def _content_parser(self) -> Callable:
        mapper = {
            "application/json": self._raw_json,
            "application/x-www-form-urlencoded": self._raw_form,
            "avro/binary": self._raw_avro,
            "text/plain": self._raw_text,
            "application/octet-stream": self._raw_bytes,
        }

        if self.content_type not in mapper:
            raise ValueError(
                f"The given Content-Type ({self.content_type!r}) is not supported for automatic content parsing yet."
            )

        return mapper[self.content_type]

    async def _raw_json(self) -> Any:
        return await self.raw.json()

    async def _raw_form(self) -> dict[str, Any]:
        form = await self.raw.json(loads=parse_qsl)
        return self._parse_multi_dict(form)

    async def _raw_avro(self) -> Any:
        data = MinosAvroProtocol.decode(await self._raw_bytes())
        schema = MinosAvroProtocol.decode_schema(await self._raw_bytes())

        type_ = AvroSchemaDecoder(schema).build()
        return AvroDataDecoder(type_).build(data)

    async def _raw_text(self) -> str:
        return await self.raw.text()

    async def _raw_bytes(self) -> bytes:
        return await self.raw.read()

    @property
    def content_type(self) -> str:
        """Get the content type.

        :return: A ``str`` value.
        """
        return self.raw.content_type

    @property
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        sentinel = object()
        return next(chain(self._raw_url_params, self._raw_query_params), sentinel) is not sentinel

    async def _params(self, type_: Optional[Union[type, str]] = None, **kwargs) -> dict[str, Any]:
        data = self._parse_multi_dict(chain(self._raw_url_params, self._raw_query_params))
        return self._build(data, type_)

    async def url_params(self, type_: Optional[Union[type, str]] = None, **kwargs) -> Any:
        """Get the url params.

        :param type_: Optional ``type`` or ``str`` (classname) that defines the request content type.
        :param kwargs: Additional named arguments.
        :return: A dictionary instance.
        """
        if not self.has_url_params:
            raise NotHasParamsException(f"{self!r} has not url params.")

        data = self._parse_multi_dict(self._raw_url_params)
        return self._build(data, type_)

    @property
    def has_url_params(self) -> bool:
        """Check if the request has url params.

        :return: ``True`` if it has url params or ``False`` otherwise.
        """
        sentinel = object()
        return next(iter(self._raw_url_params), sentinel) is not sentinel

    @property
    def _raw_url_params(self):
        return self.raw.rel_url.query.items()  # pragma: no cover

    async def query_params(self, type_: Optional[Union[type, str]] = None, **kwargs) -> Any:
        """Get the query params.


        :param type_: Optional ``type`` or ``str`` (classname) that defines the request content type.
        :param kwargs: Additional named arguments.
        :return: A dictionary instance.
        """
        if not self.has_query_params:
            raise NotHasParamsException(f"{self!r} has not query params.")

        data = self._parse_multi_dict(self._raw_query_params)
        return self._build(data, type_)

    @property
    def has_query_params(self) -> bool:
        # noinspection GrazieInspection
        """Check if the request has query params.

        :return: ``True`` if it has query params or ``False`` otherwise.
        """
        sentinel = object()
        return next(iter(self._raw_query_params), sentinel) is not sentinel

    @property
    def _raw_query_params(self):
        return self.raw.match_info.items()  # pragma: no cover

    @staticmethod
    def _build(data: Any, type_: Union[type, str]) -> Any:
        if type_ is None:
            return data

        if isinstance(type_, str):
            type_ = import_module(type_)

        return AvroDataDecoder(type_).build(data)

    @staticmethod
    def _parse_multi_dict(raw: Iterable[str, Any]) -> dict[str, Any]:
        args = defaultdict(list)
        for k, v in raw:
            args[k].append(v)
        return {k: v if len(v) > 1 else v[0] for k, v in args.items()}


class RestResponse(Response):
    """Rest Response class."""

    def __init__(self, *args, content_type: str = "application/json"):
        super().__init__(*args)
        self.content_type = content_type

    @classmethod
    def from_response(cls, response: Optional[Response]) -> RestResponse:
        """Build a new ``RestRequest`` from another response.

        :param response: The base response.
        :return: A ``RestResponse`` instance.
        """
        if isinstance(response, RestResponse):
            return response

        if response is None:
            return RestResponse()

        return RestResponse(response._data)

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Optional[bytes]:
        """Raw response content.

        :param kwargs: Additional named arguments.
        :return: The raw content as a ``bytes`` instance.
        """
        if not self.has_content:
            return None
        return await self._content_parser()

    @cached_property
    def _content_parser(self) -> Callable[[], Awaitable[bytes]]:
        mapper = {
            "application/json": self._raw_json,
            "application/x-www-form-urlencoded": self._raw_form,
            "avro/binary": self._raw_avro,
            "text/plain": self._raw_text,
            "application/octet-stream": self._raw_bytes,
        }

        if self.content_type not in mapper:
            return self._raw_bytes

        return mapper[self.content_type]

    async def _raw_json(self) -> bytes:
        return orjson.dumps(self._raw_data)

    async def _raw_form(self) -> bytes:
        return urlencode(self._raw_data).encode()

    async def _raw_avro(self) -> bytes:
        type_ = TypeHintBuilder(self._data).build()
        schema = AvroSchemaEncoder(type_).build()
        data = AvroDataEncoder(self._data).build()

        return MinosAvroProtocol.encode(data, schema)

    async def _raw_text(self) -> bytes:
        if not isinstance(self._raw_data, str):
            raise ValueError(
                f"Given 'Content-Type' ({self.content_type!r}) is not supported for the given data: {self._raw_data!r}."
            )
        return self._raw_data.encode()

    async def _raw_bytes(self) -> bytes:
        if not isinstance(self._raw_data, bytes):
            raise ValueError(
                f"Given 'Content-Type' ({self.content_type!r}) is not supported for the given data: {self._raw_data!r}."
            )
        return self._raw_data

    # noinspection PyUnusedLocal
    @cached_property
    def _raw_data(self) -> Any:
        """Raw response content.

        :return: A list of raw items.
        """
        return AvroDataEncoder(self._data).build()


class RestResponseException(ResponseException):
    """Rest Response Exception class."""



from __future__ import (
    annotations,
)

from asyncio import (
    Queue,
)
from collections.abc import (
    Iterable,
)

from ..messages import (
    BrokerMessage,
)
from .abc import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)


class InMemoryBrokerSubscriber(BrokerSubscriber):
    """In Memory Broker Subscriber class."""

    _queue: Queue[BrokerMessage]

    def __init__(self, topics: Iterable[str], messages: Iterable[BrokerMessage] = tuple(), **kwargs):
        super().__init__(topics, **kwargs)
        self._queue = Queue()
        for message in messages:
            self._queue.put_nowait(message)

    def add_message(self, message: BrokerMessage) -> None:
        """Add a message to the subscriber.

        :param message: The message to be added.
        :return: This method does not return anything.
        """
        self._queue.put_nowait(message)

    async def _receive(self) -> BrokerMessage:
        return await self._queue.get()


class InMemoryBrokerSubscriberBuilder(BrokerSubscriberBuilder):
    """In Memory Broker Subscriber Builder class."""

    def with_messages(self, messages: Iterable[BrokerMessage]) -> InMemoryBrokerSubscriberBuilder:
        """Set messages.

        :param messages: The topics to be set.
        :return: This method return the builder instance.
        """
        self.kwargs["messages"] = messages
        return self

    def build(self) -> BrokerSubscriber:
        """Build the instance.

        :return: An ``InMemoryBrokerSubscriber`` instance.
        """
        return InMemoryBrokerSubscriber(**self.kwargs)



from __future__ import (
    annotations,
)

from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from cached_property import (
    cached_property,
)

from ...requests import (
    Request,
    Response,
    ResponseException,
)
from ..messages import (
    BrokerMessage,
)


class BrokerRequest(Request):
    """Handler Request class."""

    __slots__ = "raw"

    def __init__(self, raw: BrokerMessage, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.raw = raw

    def __eq__(self, other: BrokerRequest) -> bool:
        return type(self) == type(other) and self.raw == other.raw

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self.raw!r})"

    @cached_property
    def user(self) -> Optional[UUID]:
        """
        Returns the UUID of the user making the Request.
        """
        if "user" not in self.headers:
            return None
        return UUID(self.headers["user"])

    @property
    def headers(self) -> dict[str, str]:
        """Get the headers of the request.

        :return: A dictionary in which keys are ``str`` instances and values are ``str`` instances.
        """
        return self.raw.headers

    @property
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        return True

    async def _content(self, **kwargs) -> Any:
        return self.raw.content

    @property
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        return False


class BrokerResponse(Response):
    """Handler Response class."""


class BrokerResponseException(ResponseException):
    """Handler Response Exception class."""



from __future__ import (
    annotations,
)

from asyncio import (
    iscoroutinefunction,
)
from collections.abc import (
    Iterable,
)
from datetime import (
    timedelta,
)
from typing import (
    Union,
)

from ..callables import (
    Checker,
    CheckerMeta,
    CheckerWrapper,
    Handler,
    HandlerMeta,
)


class CheckDecorator:
    """Enroute Check Decorator class."""

    def __init__(self, handler_meta: HandlerMeta, max_attempts: int = 10, delay: Union[float, timedelta] = 0.1):
        if isinstance(delay, timedelta):
            delay = delay.total_seconds()

        self.max_attempts = max_attempts
        self.delay = delay

        self.handler_meta = handler_meta

    def __call__(self, func: Union[CheckerWrapper, Checker]) -> CheckerWrapper:
        if isinstance(func, CheckerWrapper):
            func = func.meta.func

        if iscoroutinefunction(func) and not iscoroutinefunction(self._handler_func):
            raise ValueError(f"{self._handler_func!r} must be a coroutine if {func!r} is a coroutine")

        meta = CheckerMeta(func, self.max_attempts, self.delay)

        self._handler_checkers.add(meta)

        return meta.wrapper

    @property
    def _handler_func(self) -> Handler:
        return self.handler_meta.func

    @property
    def _handler_checkers(self) -> set[CheckerMeta]:
        return self.handler_meta.checkers

    def __repr__(self):
        args = ", ".join(map(repr, self))
        return f"{type(self).__name__}({args})"

    def __eq__(self, other: CheckDecorator) -> bool:
        return isinstance(other, type(self)) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        yield from (
            self.handler_meta,
            self.max_attempts,
            self.delay,
        )



import warnings
from typing import (
    Any,
    Optional,
)
from uuid import (
    UUID,
)

from .abc import (
    REQUEST_USER_CONTEXT_VAR,
    Request,
)

sentinel = object()


class InMemoryRequest(Request):
    """In Memory Request class."""

    def __init__(
        self, content: Any = sentinel, params: dict[str, Any] = sentinel, user: Optional[UUID] = None, *args, **kwargs
    ):
        super().__init__(*args, **kwargs)

        if (context_user := REQUEST_USER_CONTEXT_VAR.get()) is not None:
            if user is not None:
                warnings.warn("The `user` argument will be ignored in favor of the `user` ContextVar", RuntimeWarning)
            user = context_user

        self._content_value = content
        self._params_value = params
        self._user = user

    @property
    def user(self) -> Optional[UUID]:
        """For testing purposes"""
        return self._user

    @property
    def has_content(self) -> bool:
        """Check if the request has content.

        :return: ``True`` if it has content or ``False`` otherwise.
        """
        return self._content_value is not sentinel

    async def _content(self, **kwargs) -> Any:
        return self._content_value

    @property
    def has_params(self) -> bool:
        """Check if the request has params.

        :return: ``True`` if it has params or ``False`` otherwise.
        """
        return self._params_value is not sentinel

    async def _params(self, **kwargs) -> dict[str, Any]:
        return self._params_value

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, type(self))
            and self._content_value == other._content_value
            and self._params_value == other._params_value
            and self._user == other._user
        )

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._content_value!r}, {self._params_value!r}, {self._user!r})"



from asyncio import (
    CancelledError,
    TimeoutError,
    create_task,
    wait_for,
)
from collections.abc import (
    Awaitable,
    Iterable,
)
from contextlib import (
    suppress,
)
from typing import (
    Any,
    NoReturn,
    Optional,
)

from minos.common import (
    MinosConfig,
)

from ...messages import (
    BrokerMessage,
)
from ..abc import (
    BrokerSubscriber,
    BrokerSubscriberBuilder,
)
from .queues import (
    BrokerSubscriberQueue,
    BrokerSubscriberQueueBuilder,
)


class QueuedBrokerSubscriber(BrokerSubscriber):
    """Queued Broker Subscriber class."""

    impl: BrokerSubscriber
    queue: BrokerSubscriberQueue

    def __init__(self, impl: BrokerSubscriber, queue: BrokerSubscriberQueue, **kwargs):
        super().__init__(kwargs.pop("topics", impl.topics), **kwargs)
        if self.topics != impl.topics or self.topics != queue.topics:
            raise ValueError("The topics from the impl and queue must be equal")

        self.impl = impl
        self.queue = queue

        self._run_task = None

    async def _setup(self) -> None:
        await super()._setup()
        await self.queue.setup()
        await self.impl.setup()
        await self._start_run()

    async def _destroy(self) -> None:
        await self._stop_run()
        await self.impl.destroy()
        await self.queue.destroy()
        await super()._destroy()

    async def _start_run(self):
        if self._run_task is None:
            self._run_task = create_task(self._run())

    async def _stop_run(self):
        if self._run_task is not None:
            self._run_task.cancel()
            with suppress(TimeoutError, CancelledError):
                await wait_for(self._run_task, 0.5)
            self._run_task = None

    async def _run(self) -> NoReturn:
        async for message in self.impl:
            await self.queue.enqueue(message)

    def _receive(self) -> Awaitable[BrokerMessage]:
        return self.queue.dequeue()


class QueuedBrokerSubscriberBuilder(BrokerSubscriberBuilder):
    """Queued Broker Subscriber Publisher class."""

    def __init__(
        self, *args, impl_builder: BrokerSubscriberBuilder, queue_builder: BrokerSubscriberQueueBuilder, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.impl_builder = impl_builder
        self.queue_builder = queue_builder

    def with_config(self, config: MinosConfig) -> BrokerSubscriberBuilder:
        """Set config.

        :param config: The config to be set.
        :return: This method return the builder instance.
        """
        self.impl_builder.with_config(config)
        self.queue_builder.with_config(config)
        return super().with_config(config)

    def with_kwargs(self, kwargs: dict[str, Any]) -> BrokerSubscriberBuilder:
        """Set kwargs.

        :param kwargs: The kwargs to be set.
        :return: This method return the builder instance.
        """
        self.impl_builder.with_kwargs(kwargs)
        self.queue_builder.with_kwargs(kwargs)
        return super().with_kwargs(kwargs)

    def with_topics(self, topics: Iterable[str]) -> BrokerSubscriberBuilder:
        """Set topics.

        :param topics: The topics to be set.
        :return: This method return the builder instance.
        """
        topics = set(topics)
        self.impl_builder.with_topics(topics)
        self.queue_builder.with_topics(topics)
        return super().with_topics(topics)

    def with_group_id(self, group_id: Optional[str]) -> BrokerSubscriberBuilder:
        """Set group_id.

        :param group_id: The group_id to be set.
        :return: This method return the builder instance.
        """
        self.impl_builder.with_group_id(group_id)
        return super().with_group_id(group_id)

    def with_remove_topics_on_destroy(self, remove_topics_on_destroy: bool) -> BrokerSubscriberBuilder:
        """Set remove_topics_on_destroy.

        :param remove_topics_on_destroy: The remove_topics_on_destroy flag to be set.
        :return: This method return the builder instance.
        """
        self.impl_builder.with_remove_topics_on_destroy(remove_topics_on_destroy)
        return super().with_remove_topics_on_destroy(remove_topics_on_destroy)

    def build(self) -> BrokerSubscriber:
        """Build the instance.

        :return: A ``QueuedBrokerSubscriber`` instance.
        """
        impl = self.impl_builder.build()
        queue = self.queue_builder.build()
        return QueuedBrokerSubscriber(impl=impl, queue=queue, **self.kwargs)



from __future__ import (
    annotations,
)

import logging
from contextvars import (
    Token,
)
from typing import (
    Any,
    AsyncContextManager,
    Optional,
)

from minos.common import (
    MinosConfig,
    MinosPool,
)

from .clients import (
    BrokerClient,
)
from .messages import (
    REQUEST_REPLY_TOPIC_CONTEXT_VAR,
)

logger = logging.getLogger(__name__)


class BrokerClientPool(MinosPool):
    """Broker Client Pool class."""

    def __init__(
        self, instance_kwargs: dict[str, Any], maxsize: int = 5, recycle: Optional[int] = 3600, *args, **kwargs
    ):
        super().__init__(maxsize=maxsize, recycle=recycle, *args, **kwargs)
        self._instance_kwargs = instance_kwargs

    @classmethod
    def _from_config(cls, config: MinosConfig, **kwargs) -> BrokerClientPool:
        return cls(kwargs | {"config": config})

    async def _create_instance(self) -> BrokerClient:
        instance = BrokerClient.from_config(**self._instance_kwargs)
        await instance.setup()
        return instance

    async def _destroy_instance(self, instance: BrokerClient):
        await instance.destroy()

    def acquire(self, *args, **kwargs) -> AsyncContextManager:
        """Acquire a new instance wrapped on an asynchronous context manager.

        :return: An asynchronous context manager.
        """
        return _ReplyTopicContextManager(super().acquire())


class _ReplyTopicContextManager:
    _token: Optional[Token]

    def __init__(self, wrapper: AsyncContextManager[BrokerClient]):
        self.wrapper = wrapper
        self._token = None

    async def __aenter__(self) -> BrokerClient:
        broker = await self.wrapper.__aenter__()
        self._token = REQUEST_REPLY_TOPIC_CONTEXT_VAR.set(broker.topic)
        return broker

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        REQUEST_REPLY_TOPIC_CONTEXT_VAR.reset(self._token)
        await self.wrapper.__aexit__(exc_type, exc_val, exc_tb)
